<?php

abstract class queryTester {
    
    //@var object db connection
    protected $_oDb;
    //@var array The data to use for the tests
    protected $_aTestData;
    //@var array The sql queries to be tested
    protected $_aQueries;
    //@var array The results of the tests
    protected $_aResults;
    //@var int The maximum time an individual query should run for
    protected $_iMaxQueryTime = 1;
    //@var float The time it took for the last run query to execute and fetch all results
    protected $_fTimeTaken;
    //@var string The query type, either Select or Update. (Any query that returns results is classed as Select while all others are classed as Update)
    protected $_sQueryType;
    
    
    /**
     *Sets the data to use for the tests
     *@param array $aTestData The data to use for the tests
     *@return object $this
     */
    public function setTestData(array $aTestData)
    {
        $this->_aTestData = $aTestData;
        return $this;
    }
    
    /**
     *Sets the queries to be tested
     *@param array $aQueries The queries to be tested
     *@return object $this
     */
    public function setQueries(array $aQueries)
    {
        $this->_aQueries = $aQueries;
        return $this;
    }
    
    /**
     *Runs the queries for the given amount of times in total and returns the timing results
     *@param int $iNumTests default 10000 The total number of queries to execute - note this will be split between the queries, so if there are 2 queries and NumTests is 500, one query may be executed 273 times and the other 227 times (for example).
     *@return array The test results
     */
    public function runTests($iNumTests=10)
    {
        if (!is_int($iNumTests)) {
            throw new Exception('queryTester::setNumTests expects argument to be of type integer, argument is of type '.gettype($iNumTests));
        }
        //prepare the results array
        $this->_initResultsArray();
        //run the tests
        for ($i=0;$i<$iNumTests;$i++) {
            //select a random query
            $mQueryKey = array_rand($this->_aQueries);
            $sQuery = $this->_aQueries[$mQueryKey];
            //select random test data
            $mDataKey = array_rand($this->_aTestData);
            $aData = $this->_aTestData[$mDataKey];
            //prepare the query
            $sSQL = $this->_prepareQuery($sQuery,$aData);
            //run the query - how long it took will be stored in $this->_fTimeTaken property
            $this->_timeQuery($sSQL);
            //store the time taken in the results
            $this->_aResults[$mQueryKey]['aPerDataResults'][$mDataKey]['iNum']++;
            $this->_aResults[$mQueryKey]['aPerDataResults'][$mDataKey]['fTotal'] += $this->_fTimeTaken;
            $this->_aResults[$mQueryKey]['aTotalResults']['iNum']++;
            $this->_aResults[$mQueryKey]['aTotalResults']['fTotal'] += $this->_fTimeTaken;
        }
        //calculate the averages
        $this->_calcAvgs();
        return $this->_aResults;
    }
    
    /**
     *Inserts data into the query to be executed. Does not escape anything.
     *@param string $sQuery SQL query formatted using sprintf conversion characters to indicate where data should be inserted
     *@param array $aData The data to be inserted into the query, indexed in the same order it should be inserted
     *@return string The query with the data inserted
     */
    protected function _prepareQuery($sQuery,$aData)
    {
        return vsprintf($sQuery, $aData);;
    }
    
    
    /**
     *Adds the average time taken to all the results
     */
    protected function _calcAvgs()
    {
        foreach ($this->_aResults as &$aQueryResults) {
            $fTotalAvg = 0;
            //for each of the test datas, calculate the average
            foreach ($aQueryResults['aPerDataResults'] as &$aDataResult) {
                if ($aDataResult['fTotal'] && $aDataResult['iNum']) {
                    $fTotalAvg += $aDataResult['fAvg'] = $aDataResult['fTotal']/$aDataResult['iNum'];
                }
            }
            //add the overall average time
            $aTotalResults = &$aQueryResults['aTotalResults'];
            if ($aTotalResults['fTotal'] && $aTotalResults['iNum']) {
                $aTotalResults['fAvg'] = $fTotalAvg/count($aQueryResults['aPerDataResults']);
            }
             
        }
    }
    
    /**
     *Times the run for a single query
     *@param string $sSQL The query to run
     *@return array Results fetched by the query
     */
    protected function _timeQuery($sSQL)
    {
        //type of query (Select or Update)
        $sQueryType = $this->getQueryType();
        //Prevent PHP timeout - 20s for script to finish running, plus the max query time
        set_time_limit(20+$this->_iMaxQueryTime);
        //timer to check how long query takes
        $fStart = microtime(true);
        
        //fetch the results using database here
        $aResults = $this->{'_runQuery_'.$sQueryType}($sSQL);
        
        //If query took too long, then exit to prevent running slow query lots of times
        $this->_fTimeTaken = $fTimeTaken = (microtime(true) - $fStart);
        if ($fTimeTaken  > $this->_iMaxQueryTime ) {
            throw new Exception("Timed out after query took $fTimeTaken seconds to execute and retrieve results for, which is over the allowed limit of $this->_iMaxQueryTime seconds. Query was: \n".$sSQL);
        }
        return $aResults;
    }
    
    
    /**
     *Runs a single query, fetching all rows the query returns from the db
     *@param string $sSQL The query to run
     *@return array Results fetched by the query
     */
    abstract protected function _runQuery_Select($sSQL);
    
    
    /**
     *Runs a single query, returning the number of rows the query affected
     *@param string $sSQL The query to run
     *@return array Array containing the number of rows the query affected
     */
    abstract protected function _runQuery_Update($sSQL);
    
    /**
     *Sets the maximum time a query should take
     *If a query takes longer than this, then an exception will be thrown, preventing any more queries being executed
     *Note that this does not restrict the actual time a query may take, just halts the script after a slow query has finished executing.
     *@param int $iSeconds The max time a query should be allowed to run for
     *@return object $this
     */
    public function setMaxQueryTime($iSeconds=1)
    {
        if (!is_int($iSeconds)) {
            throw new Exception('queryTester::setMaxQueryTime expects argument to be of type integer, argument passed is of type'.gettype($iSeconds));
        }
        $this->_iMaxQueryTime = $iSeconds;
        return $this;
    }
    
    /**
     *Tests the queries, ensuring they all give the same results
     *@param bool $bStoreSQL default false Whether the actual sql queries executed should be stored as part of the results (can be helpful in debugging if queries not giving identical results)
     *@param bool $bStoreResults default false Whether the results fetched from the db for each query should be stored as part of the results (can be helpful to var_dump and compare if queries not giving identical results)
     *@return array Associative array. Index 'errors' is an array that lists any data for which not all queries gave identical results. Index 'results' is an array with the results of the comparisons, useful for debugging.
     */
    public function checkQueries($bStoreSQL=false, $bStoreResults=false)
    {
        //array to store the results of our check
        $aResults = array();
        //array to store whether any query gave different results to other queries with the same data
        $aErrors = array();
        //type of query (Select or Update)
        $sQueryType = $this->getQueryType(); 
        //loop through the test data
        foreach ($this->_aTestData as $mDataKey => $aData) {
            $aResults[$mDataKey] = array();
            //loop through the queries
            foreach ($this->_aQueries as $mQueryKey => $sQuery) {                
                //prepare the query
                $sSql = $this->_prepareQuery($sQuery,$aData);
                //serialize the results to convert them to a string, then md5 this so that we can easily compare whether the results are the same or not
                //we use this as the key, so all queries that give the same results will be stored under the same key                
                $aResultsFromDb = $this->{'_runQuery_'.$sQueryType}($sSql);
                $sResultKey = md5(serialize($aResultsFromDb));
                //create an empty array using the result key if it doesn't already exist
                if (empty($aResults[$mDataKey][$sResultKey])) {
                    $aResults[$mDataKey][$sResultKey]=array();
                }
                //create an array to store the sql and results if wanted
                //useful for var_dumping if results aren't identical to see what the difference is
                $aResult = array('key' => $mQueryKey);
                if ($bStoreSQL) {
                    $aResult['sql'] = $sSql;
                }
                if ($bStoreResults) {
                    $aResult['results'] = $aResultsFromDb;
                }
                $aResults[$mDataKey][$sResultKey][$mQueryKey] = $aResult;
            }
            //if results were identical, then all entries should be stored under the same result key, so there will only be one entry in the array for this data
            if ( count($aResults[$mDataKey]) > 1 ) {
                $aErrors[$mDataKey] = 'Some of the queries returned non-identical result sets when using data with key '.$mDataKey;
            }
        }
        return array('errors' => $aErrors, 'results' => $aResults);
    }
    
    /**
     *Gets the type of query, for determining how to run the query (look for results or look for affected rows)
     *@return string The query type, either 'Select' or 'Update'
     */
    public function getQueryType()
    {
        if ($this->_sQueryType) {
            return $this->_sQueryType;
        }
        $sSqlStart = substr(reset($this->_aQueries), 0, 8);
        $iSelect = preg_match('/^SELECT|SHOW|DESCRIBE|EXPLAIN/i', $sSqlStart);
        if ($iSelect === false) {
            throw new Exception('error occured when trying to determine query type');
        } elseif ($iSelect) {
            return $this->_sQueryType = 'Select';
        } else {
            return $this->_sQueryType = 'Update';
        }
    }
    
    /**
     *Sets the type of query, for determining how to run the query (look for results or look for affected rows)
     *@param string The query type, either 'Select' or 'Update'
     *@return object instance
     */
    public function setQueryType($sQueryType) {
        if ($sQueryType == 'Select') {
            $this->_sQueryType = 'Select';
        } elseif ($sQueryType === 'Update') {
            $this->_sQueryType = 'Update';
        } else {
            throw new Exception("$sQueryType is not a valid value. Query Type should either be set to 'Select' for queries that return results, or 'Update' for queries that affect records.");
        }
        return $this;
    }
    
    /**
     *Sets up the results array
     */
    protected function _initResultsArray()
    {
        //array used to store the results in for each item
        $aEmptyResults = array('iNum' => 0, 'fTotal' => 0, 'fAvg' => 0);
        foreach ($this->_aQueries as $mQueryKey => $sQuery) {
            //for each of the tests, create a corresponding result entry
            $this->_aResults[$mQueryKey] = array('query' => $sQuery, 'aPerDataResults' => array(), 'aTotalResults' => $aEmptyResults);
            foreach ($this->_aTestData as $mDataKey => $mData) {
                //for each of the test datas, create a corresponding result entry under the query
                $this->_aResults[$mQueryKey]['aPerDataResults'][$mDataKey] = $aEmptyResults;
            }
        }
    }
    
    /**
     *Formats the results as CSV
     *@param string $sPath Path to save CSV file to Use null for output, empty string for return csv as string
     *@return string EMpty string is returned when outputting CSV or saving it to a file. Otherwise a string with the results in CSV format is returned
     */
    public function getResultsAsCSV($sPath=null)
    {
        if (empty($this->_aResults)) {
            throw new Exception('No results to create CSV for');
        }
        //initialise an array to hold the rows for the CSV
        $aCSV = array();
        //loop through the data. Each data entry will have a row, while the query results form the columns
        foreach ($this->_aTestData as $mDataKey => $aData) {
            //create a row for this data
            $aCSV[$mDataKey] = array($mDataKey);
            foreach ($this->_aResults as $mQueryKey => $aResult) {
                //add the results for this query to the data row
                $aCSV[$mDataKey]=array_merge($aCSV[$mDataKey], array_values($aResult["aPerDataResults"][$mDataKey]));
            }
        }
        //Now create the query headings and the total row
        //2nd row, column headings with num, total, avg
        array_unshift($aCSV, array('data'));
        //1st row, column headings with query names
        array_unshift($aCSV, array(''));
        //last row, totals
        $aLastRow = array('total');
        foreach ($this->_aResults as $mQueryKey => $aResult) {
            //3 columns for each query - num, total, and avg
            $aCSV[1]=array_merge($aCSV[1], array('num','total','avg'));
            //query title only takes up 1 col so following 2 must be blank
            $aCSV[0]=array_merge($aCSV[0], array($mQueryKey,'',''));
            //totals row
            $aLastRow=array_merge($aLastRow, array_values($aResult["aTotalResults"]));
        }
        //add the last row to the array
        $aCSV[] = $aLastRow;
        
        
        //Create the actual CSV
        //http://stackoverflow.com/questions/6325613/escaping-for-csv#answer-31118307
        if ($sPath) { //output to file
            $fp = fopen($sPath, 'w');
        } elseif ($sPath === '') { //capture output to return
            $fp = fopen('php://temp', 'w+');
        } else { //output as response
            header('Content-Type: text/csv; charset=utf-8');
            header('Content-Disposition: attachment; filename=query-benchmark-results.csv');
            $fp = fopen('php://output', 'w');
        }
        
        //write the CSV
        foreach ($aCSV as $aRow) {
            // Add row to CSV buffer
            fputcsv($fp, $aRow);
        }
        $sCSV='';
        //if we want to return the CSV as a string
        if ($sPath === '') {
            //Set the pointer back to the start
            rewind($fp);
            //Fetch the contents of our CSV
            $sCSV = stream_get_contents($fp);
        }
        //Close our pointer and free up memory and /tmp space
        fclose($fp); 
        return $sCSV;
    }
}
